"""Defines a logger which logs JSON lines to a file."""

import json
from dataclasses import asdict
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Literal

from omegaconf import DictConfig, OmegaConf
from torch import Tensor

from mlfab.task.logger import LoggerImpl, LogLine


def get_json_value(value: Any) -> Any:  # noqa: ANN401
    if isinstance(value, Tensor):
        value = value.detach().float().cpu().item()
    return value


class JsonLogger(LoggerImpl):
    def __init__(
        self,
        run_directory: str | Path,
        log_file_name: str = "log.jsonl",
        git_state_name: str = "git_state.txt",
        train_code_name: str = "train_code.py",
        config_name: str = "config.yaml",
        flush_immediately: bool = False,
        open_mode: Literal["w", "a"] = "w",
        line_sep: str = "\n",
        remove_unicode_from_namespaces: bool = True,
        log_interval_seconds: float = 10.0,
    ) -> None:
        """Defines a simpler logger which logs to stdout.

        Args:
            run_directory: The directory to log to.
            log_file_name: The name of the log file.
            git_state_name: The name of the git state file.
            train_code_name: The name of the training code file.
            config_name: The name of the config file.
            flush_immediately: Whether to flush the file after every write.
            open_mode: The file open mode.
            line_sep: The line separator to use.
            remove_unicode_from_namespaces: Whether to remove unicode from
                namespaces. This is the typical behavior for namespaces that
                use ASCII art for visibility in other logs, but in the JSON
                log file should be ignored.
            log_interval_seconds: The interval between successive log lines.
        """
        super().__init__(log_interval_seconds)

        self.log_file = Path(run_directory).expanduser().resolve() / log_file_name
        self.git_state_file = Path(run_directory).expanduser().resolve() / git_state_name
        self.train_code_file = Path(run_directory).expanduser().resolve() / train_code_name
        self.config_file = Path(run_directory).expanduser().resolve() / config_name
        self.flush_immediately = flush_immediately
        self.open_mode = open_mode
        self.line_sep = line_sep
        self.remove_unicode_from_namespaces = remove_unicode_from_namespaces

        self.__fp: TextIOWrapper | None = None

    def start(self) -> None:
        self.__fp = open(self.log_file, self.open_mode)
        return super().start()

    def stop(self) -> None:
        if self.__fp is not None:
            self.__fp.close()
            self.__fp = None
        return super().stop()

    @property
    def fp(self) -> TextIOWrapper:
        assert self.__fp is not None, "Not started!"
        return self.__fp

    def get_json(self, line: LogLine) -> str:
        data: dict = {"state": asdict(line.state)}
        for log in (line.scalars, line.strings):
            for namespace, values in log.items():
                if self.remove_unicode_from_namespaces:
                    namespace = namespace.encode("ascii", errors="ignore").decode("ascii").strip()
                if namespace not in data:
                    data[namespace] = {}
                for k, v in values.items():
                    data[namespace][k] = get_json_value(v)
        return json.dumps(data)

    def write(self, line: LogLine) -> None:
        self.fp.write(self.get_json(line))
        self.fp.write(self.line_sep)
        if self.flush_immediately:
            self.fp.flush()

    def log_git_state(self, git_state: str) -> None:
        with open(self.git_state_file, "w") as f:
            f.write(git_state)

    def log_training_code(self, training_code: str) -> None:
        with open(self.train_code_file, "w") as f:
            f.write(training_code)

    def log_config(self, config: DictConfig) -> None:
        OmegaConf.save(config, self.config_file)



"""Defines a mixin for incorporating some logging functionality."""

from dataclasses import dataclass
from types import TracebackType
from typing import Callable, Generic, Self, Sequence, TypeVar

from torch import Tensor

from mlfab.core.conf import Device as BaseDeviceConfig, field
from mlfab.core.state import State
from mlfab.task.base import BaseConfig, BaseTask
from mlfab.task.logger import ChannelSelectMode, Logger, LoggerImpl, Number
from mlfab.task.loggers.stdout import StdoutLogger


@dataclass
class LoggerConfig(BaseConfig):
    device: BaseDeviceConfig = field(BaseDeviceConfig(), help="Device configuration")


Config = TypeVar("Config", bound=LoggerConfig)


class LoggerMixin(BaseTask[Config], Generic[Config]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)

        self.logger = Logger()

    def add_logger(self, *logger: LoggerImpl) -> None:
        self.logger.add_logger(*logger)

    def set_loggers(self) -> None:
        self.add_logger(StdoutLogger())

    def write_logs(self, state: State) -> None:
        self.logger.write(state)

    def log_scalar(self, key: str, value: Callable[[], Number] | Number, *, namespace: str | None = None) -> None:
        self.logger.log_scalar(key, value, namespace=namespace)

    def log_string(self, key: str, value: Callable[[], str] | str, *, namespace: str | None = None) -> None:
        self.logger.log_string(key, value, namespace=namespace)

    def log_image(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        keep_resolution: bool = False,
    ) -> None:
        self.logger.log_image(
            key,
            value,
            namespace=namespace,
            keep_resolution=keep_resolution,
        )

    def log_labeled_image(
        self,
        key: str,
        value: Callable[[], tuple[Tensor, str]] | tuple[Tensor, str],
        *,
        namespace: str | None = None,
        max_line_length: int | None = None,
        keep_resolution: bool = False,
        centered: bool = True,
    ) -> None:
        self.logger.log_labeled_image(
            key,
            value,
            namespace=namespace,
            max_line_length=max_line_length,
            keep_resolution=keep_resolution,
            centered=centered,
        )

    def log_images(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        keep_resolution: bool = False,
        max_images: int | None = None,
        sep: int = 0,
    ) -> None:
        self.logger.log_images(
            key,
            value,
            namespace=namespace,
            keep_resolution=keep_resolution,
            max_images=max_images,
            sep=sep,
        )

    def log_labeled_images(
        self,
        key: str,
        value: Callable[[], tuple[Tensor, Sequence[str]]] | tuple[Tensor, Sequence[str]],
        *,
        namespace: str | None = None,
        max_line_length: int | None = None,
        keep_resolution: bool = False,
        max_images: int | None = None,
        sep: int = 0,
        centered: bool = True,
    ) -> None:
        self.logger.log_labeled_images(
            key,
            value,
            namespace=namespace,
            max_line_length=max_line_length,
            keep_resolution=keep_resolution,
            max_images=max_images,
            sep=sep,
            centered=centered,
        )

    def log_audio(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        sample_rate: int = 44100,
        log_spec: bool = True,
        n_fft_ms: float = 32.0,
        hop_length_ms: float | None = None,
        channel_select_mode: ChannelSelectMode = "first",
        keep_resolution: bool = False,
    ) -> None:
        self.logger.log_audio(
            key,
            value,
            namespace=namespace,
            sample_rate=sample_rate,
            log_spec=log_spec,
            n_fft_ms=n_fft_ms,
            hop_length_ms=hop_length_ms,
            channel_select_mode=channel_select_mode,
            keep_resolution=keep_resolution,
        )

    def log_audios(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        sep_ms: float = 0.0,
        max_audios: int | None = None,
        sample_rate: int = 44100,
        log_spec: bool = True,
        n_fft_ms: float = 32.0,
        hop_length_ms: float | None = None,
        channel_select_mode: ChannelSelectMode = "first",
        spec_sep: int = 0,
        keep_resolution: bool = False,
    ) -> None:
        self.logger.log_audios(
            key,
            value,
            namespace=namespace,
            sep_ms=sep_ms,
            max_audios=max_audios,
            sample_rate=sample_rate,
            log_spec=log_spec,
            n_fft_ms=n_fft_ms,
            hop_length_ms=hop_length_ms,
            channel_select_mode=channel_select_mode,
            spec_sep=spec_sep,
            keep_resolution=keep_resolution,
        )

    def log_spectrogram(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        sample_rate: int = 44100,
        n_fft_ms: float = 32.0,
        hop_length_ms: float | None = None,
        channel_select_mode: ChannelSelectMode = "first",
        keep_resolution: bool = False,
    ) -> None:
        self.logger.log_spectrogram(
            key,
            value,
            namespace=namespace,
            sample_rate=sample_rate,
            n_fft_ms=n_fft_ms,
            hop_length_ms=hop_length_ms,
            channel_select_mode=channel_select_mode,
            keep_resolution=keep_resolution,
        )

    def log_spectrograms(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        max_audios: int | None = None,
        sample_rate: int = 44100,
        n_fft_ms: float = 32.0,
        hop_length_ms: float | None = None,
        channel_select_mode: ChannelSelectMode = "first",
        spec_sep: int = 0,
        keep_resolution: bool = False,
    ) -> None:
        self.logger.log_spectrograms(
            key,
            value,
            namespace=namespace,
            max_audios=max_audios,
            sample_rate=sample_rate,
            n_fft_ms=n_fft_ms,
            hop_length_ms=hop_length_ms,
            channel_select_mode=channel_select_mode,
            spec_sep=spec_sep,
            keep_resolution=keep_resolution,
        )

    def log_video(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        fps: int | None = None,
        length: float | None = None,
    ) -> None:
        self.logger.log_video(
            key,
            value,
            namespace=namespace,
            fps=fps,
            length=length,
        )

    def log_videos(
        self,
        key: str,
        value: Callable[[], Tensor | list[Tensor]] | Tensor | list[Tensor],
        *,
        namespace: str | None = None,
        max_videos: int | None = None,
        sep: int = 0,
        fps: int | None = None,
        length: int | None = None,
    ) -> None:
        self.logger.log_videos(
            key,
            value,
            namespace=namespace,
            max_videos=max_videos,
            sep=sep,
            fps=fps,
            length=length,
        )

    def log_histogram(self, key: str, value: Callable[[], Tensor] | Tensor, *, namespace: str | None = None) -> None:
        self.logger.log_histogram(key, value, namespace=namespace)

    def log_point_cloud(
        self,
        key: str,
        value: Callable[[], Tensor] | Tensor,
        *,
        namespace: str | None = None,
        max_points: int = 1000,
        colors: Callable[[], Tensor] | Tensor | None = None,
    ) -> None:
        self.logger.log_point_cloud(
            key,
            value,
            namespace=namespace,
            max_points=max_points,
            colors=colors,
        )

    def __enter__(self) -> Self:
        self.logger.__enter__()
        return self

    def __exit__(self, t: type[BaseException] | None, e: BaseException | None, tr: TracebackType | None) -> None:
        self.logger.__exit__(t, e, tr)
        return super().__exit__(t, e, tr)



"""Defines a logger that shows a pop-up using Curses."""

import bdb
import curses
import datetime
import functools
import logging
import sys
from collections import deque
from typing import TYPE_CHECKING, Any, Deque, TextIO

from torch import Tensor
from wcwidth import wcswidth

from mlfab.task.logger import LoggerImpl, LogLine
from mlfab.utils.experiments import ToastKind, Toasts
from mlfab.utils.text import format_timedelta

if TYPE_CHECKING:
    from _curses import _CursesWindow


def format_number(value: int | float, precision: int) -> str:
    if isinstance(value, int):
        return str(value)
    return f"{value:.{precision}g}"


def as_str(value: Any, precision: int) -> str:  # noqa: ANN401
    if isinstance(value, str):
        return f'"{value}"'
    if isinstance(value, Tensor):
        value = value.detach().float().cpu().item()
    if isinstance(value, (int, float)):
        return format_number(value, precision)
    raise TypeError(f"Unexpected log type: {type(value)}")


class CursesLogger(LoggerImpl):
    def __init__(
        self,
        debug: bool = False,
        debug_fp: TextIO = sys.stdout,
        precision: int = 4,
        log_timers: bool = True,
        log_perf: bool = False,
        log_optim: bool = False,
        max_toasts: int = 10,
        log_interval_seconds: float = 1.0,
    ) -> None:
        """Defines a logger which shows a pop-up using Curses.

        Args:
            debug: Whether to enable debugging mode. If debugging, the
                terminal will not be modified; instead, we just print to stdout
                normally.
            debug_fp: The file to write debug logs to.
            precision: The integer precision to use when logging scalars.
            log_timers: Whether to log timers.
            log_perf: Whether to log performance metrics.
            log_optim: Whether to log optimizer parameters.
            max_toasts: The maximum number of toasts to display.
            log_interval_seconds: The interval between successive log lines.
        """
        super().__init__(log_interval_seconds)

        self.debug = debug
        self.debug_fp = debug_fp
        self.log_timers = log_timers
        self.log_perf = log_perf
        self.log_optim = log_optim
        self.precision = precision

        self.persistent_toast_queue: Deque[tuple[str, ToastKind]] = deque()
        self.temporary_toast_queue: Deque[tuple[str, ToastKind]] = deque(maxlen=max_toasts)

        Toasts.register_callback("error", functools.partial(self.handle_toast, persistent=False, kind="error"))
        Toasts.register_callback("warning", functools.partial(self.handle_toast, persistent=False, kind="warning"))
        Toasts.register_callback("info", functools.partial(self.handle_toast, persistent=False, kind="info"))
        Toasts.register_callback("status", functools.partial(self.handle_toast, persistent=True, kind="status"))

        self.__stdscr: "_CursesWindow | None" = None

    def _get_window(self) -> "_CursesWindow":
        stdscr = curses.initscr()
        curses.noecho()
        curses.cbreak()
        curses.curs_set(0)
        stdscr.keypad(True)
        if curses.has_colors():
            curses.start_color()
        return stdscr

    def start(self) -> None:
        if self.__stdscr is None:
            self.__stdscr = self._get_window()
            curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
            curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
            curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
            curses.init_pair(4, curses.COLOR_CYAN, curses.COLOR_BLACK)
            curses.init_pair(5, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
        return super().start()

    def stop(self) -> None:
        if self.__stdscr is not None:
            self.__stdscr.keypad(False)
            curses.curs_set(1)
            curses.echo()
            curses.nocbreak()
            curses.endwin()
            self.__stdscr = None
        sys.stdout.write("")
        sys.stdout.flush()
        return super().stop()

    def handle_toast(self, msg: str, persistent: bool, kind: ToastKind) -> None:
        if persistent:
            self.persistent_toast_queue.append((msg, kind))
        else:
            self.temporary_toast_queue.append((msg, kind))

    @property
    def stdscr(self) -> "_CursesWindow":
        assert self.__stdscr is not None, "Not started!"
        return self.__stdscr

    def write_state_window(self, line: LogLine, window: "_CursesWindow") -> None:
        elapsed_time = format_timedelta(datetime.timedelta(seconds=line.state.elapsed_time_s), short=True)
        state_info = {
            "Steps": f"{line.state.num_steps}",
            "Samples": f"{line.state.num_samples}",
            "Elapsed Time": f"{elapsed_time}",
        }
        if line.state.num_epochs > 0:
            state_info["Epochs"] = f"{line.state.num_epochs}"
        max_width = max(wcswidth(k) + wcswidth(v) for k, v in state_info.items()) + 2

        try:
            window.resize(len(state_info) + 3, max_width + 4)
        except Exception:
            return
        window.border()
        window.addstr(1, 2, "Phase")
        ph = line.state.phase
        window.addstr(1, max_width - wcswidth(ph) + 2, ph, curses.color_pair(2 if ph == "train" else 3))
        for i, (k, v) in enumerate(state_info.items()):
            window.addstr(i + 2, 2, k)
            window.addstr(i + 2, max_width - wcswidth(v) + 2, v, curses.color_pair(4))

    def write_log_window(self, line: LogLine, window: "_CursesWindow") -> None:
        namespace_to_lines: dict[str, dict[str, str]] = {}
        num_lines = 0
        max_width = 0
        for log in (line.scalars, line.strings):
            for namespace, values in log.items():
                if not self.log_timers and namespace.startswith("⏰"):
                    continue
                if not self.log_perf and namespace.startswith("🔧"):
                    continue
                if not self.log_optim and namespace.startswith("📉"):
                    continue
                if namespace not in namespace_to_lines:
                    namespace_to_lines[namespace] = {}
                num_lines += 2 if num_lines > 0 else 1
                max_width = max(max_width, wcswidth(namespace) + 3)
                for k, v in values.items():
                    v_str = as_str(v, self.precision)
                    namespace_to_lines[namespace][k] = v_str
                    num_lines += 1
                    max_width = max(max_width, wcswidth(k) + wcswidth(v_str) + 2)

        try:
            window.resize(num_lines + 2, max_width + 6)
        except Exception:
            return
        window.border()
        line = 1
        for namespace, lines in sorted(namespace_to_lines.items()):
            if line > 1:
                line += 1
            window.addstr(line, 2, namespace, curses.color_pair(4) | curses.A_BOLD)
            line += 1
            for k, v in lines.items():
                window.addstr(line, 2, k)
                window.addstr(line, max_width - wcswidth(v) + 4, v, curses.color_pair(5))
                line += 1

    def write_queue_window(self, q: Deque[tuple[str, ToastKind]], window: "_CursesWindow", width: int) -> None:
        if not q:
            return

        def clip_left(s: str, length: int) -> str:
            s = s.replace("\n", " ")
            return s if len(s) <= length else "..." + s[-length + 3 :]

        def get_color(kind: ToastKind) -> int:
            match kind:
                case "error":
                    return 1
                case "warning":
                    return 3
                case "info":
                    return 4
                case "status":
                    return 2
                case _:
                    return 5

        width = max(len(clip_left(msg, width - 4)) for msg, _ in q) + 4
        try:
            window.resize(len(q) + 2, width)
        except Exception:
            return
        window.border()
        for i, (msg, kind) in enumerate(reversed(q)):
            window.addstr(i + 1, 2, clip_left(msg, width - 4), curses.color_pair(get_color(kind)))

    def write(self, line: LogLine) -> None:
        try:
            self.stdscr.clear()
            _, win_width = self.stdscr.getmaxyx()

            # Writes the state window.
            try:
                state_window = self.stdscr.subwin(0, 0)
            except Exception:
                self.stdscr.refresh()
                return
            self.write_state_window(line, state_window)
            sw_height, sw_width = state_window.getmaxyx()

            # Writes the log window.
            try:
                log_window = self.stdscr.subwin(0, sw_width)
            except Exception:
                self.stdscr.refresh()
                return
            self.write_log_window(line, log_window)
            lw_height, _ = log_window.getmaxyx()

            # Writes the persistent window.
            start_y, start_x = max(sw_height, lw_height), 0
            if self.persistent_toast_queue:
                try:
                    persistent_window = self.stdscr.subwin(start_y, start_x)
                except Exception:
                    self.stdscr.refresh()
                    return
                self.write_queue_window(self.persistent_toast_queue, persistent_window, min(win_width, 100))
                pw_height, _ = persistent_window.getmaxyx()
                start_y += pw_height

            # Writes the warning window.
            if self.temporary_toast_queue:
                try:
                    temporary_window = self.stdscr.subwin(start_y, start_x)
                except Exception:
                    self.stdscr.refresh()
                    return
                self.write_queue_window(self.temporary_toast_queue, temporary_window, min(win_width, 100))
                tw_height, _ = temporary_window.getmaxyx()
                start_y += tw_height

            self.stdscr.refresh()

        except (KeyboardInterrupt, bdb.BdbQuit):
            raise

        except Exception:
            logging.exception("Error while writing log line")
